use crate::{brian::InterpretError, lexer::Token};
use std::{fmt::Display, io, ops::Range};

#[derive(Debug)]
pub struct Error {
    pub kind: ErrorKind,
    pub span: Range<usize>,
}

#[derive(Debug)]
pub enum ErrorKind {
    UnexpectedEof,
    UnexpectedToken(Token),
    UnknownVariable(String),
    MeloVariable(String),
    TopLevelEnough,
    NonExitingRlyeh(i32),
    MissingLhs,
    Brian(InterpretError),
    Io(io::Error),
}

impl Error {
    pub fn new(kind: ErrorKind, span: Range<usize>) -> Self {
        Self { kind, span }
    }

    /// Create an UnexpectedEof error, where the EOF occurs at the
    /// given index in the file.
    pub fn unexpected_eof(index: usize) -> Self {
        Self::new(ErrorKind::UnexpectedEof, index..index)
    }
}

impl Display for Error {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(
            f,
            "Error at range {}-{}: {}",
            self.span.start, self.span.end, self.kind
        )
    }
}
impl std::error::Error for Error {}

impl Display for ErrorKind {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            ErrorKind::UnexpectedEof => write!(f, "unexpected end of file"),
            ErrorKind::UnexpectedToken(Token::Melo) => write!(f, "unexpected marten"),
            ErrorKind::UnexpectedToken(token) => write!(f, "unexpected token {:?}", token),
            ErrorKind::UnknownVariable(name) => write!(f, "unknown identifier \"{}\"", name),
            ErrorKind::MeloVariable(name) => write!(f, "banned variable \"{}\"", name),
            ErrorKind::TopLevelEnough => write!(f, "can only `enough` out of a loop"),
            &ErrorKind::NonExitingRlyeh(code) => write!(f, "program exited with code {code}"),
            ErrorKind::Brian(err) => write!(f, "brainfuck error: {}", err),
            // TODO: give concrete numbers here.
            ErrorKind::MissingLhs => write!(f, "missing expression before binary operation"),
            ErrorKind::Io(err) => write!(f, "I/O error: {}", err),
        }
    }
}

impl From<io::Error> for ErrorKind {
    fn from(e: io::Error) -> Self {
        Self::Io(e)
    }
}
